/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "fetch_stage.hh"
#include "frame_source.hh"
#include "driver_table.hh"
#include "source_histo.hh"

#include <iostream>
#include <sstream>
#include <stdexcept>

using namespace sends;
using namespace std;

//======================================  Construct fetch stage
fetch_stage::fetch_stage(int nThread, int nTrans)
  : seda_stage(nThread, nTrans)
{
}

//======================================  Destroy fetch stage
fetch_stage::~fetch_stage(void) {
}

//======================================  Return stage name.
const char*
fetch_stage::stage_name(void) const {
    return "fetch";
}

//======================================  Fetch data.
void
fetch_stage::filter(transaction* trans) {
    trans->set_stage(400);

    //----------------------------------  Try to process
    try {

	//------------------------------  Decide on processing 
	switch (trans->command()) {
	case transaction::kGetData:
	case transaction::kGetOnlineData:
	case transaction::kGetSourceList:
	case transaction::kGetSourceData:
	    find_data(trans);
	default:
	    break;
	}

	//------------------------------  Pass data on to next stage
	trans->set_stage(410);
	put_output(trans);

    //----------------------------------  Catch exceptions.  
    } catch (exception& e) {
	trans->set_stage(420);
	if (debug(dlvl_detail)) trans->dump(cout);
	reject(trans, e.what());
    }
}

//======================================  Count unresolved channels.
static int
count_nready(const transaction* trans) {
    int N = trans->getNChan();
    int nChan(0);
    for (int i=0; i<N; ++i) {
	if (!trans->getChanReady(i)) nChan++;
    }
    return nChan;
}

//======================================  Loop over channels read data.
void
fetch_stage::find_data(transaction* trans) {

    //----------------------------------  Set up for smooth running
    index_type N = trans->getNChan();
    source_histo sl;
    bool trans_online = (trans->command() == transaction::kGetOnlineData);
    int iter_count = 0;

    //----------------------------------  Iterate until everything is resolved
    debug_msg(dlvl_meta, trans->ident(), "Loop over sources.");
    while (count_nready(trans)) {
	if (!trans->test_connect()) {
	    throw runtime_error("Client closed connection");
	}
	iter_count++;

	//------------------------------  Histogram frame source types
	//
	//   Loop over channel names and source strings, histogramming
	//   the number of channels available from each source.
	if (debug(dlvl_meta)) cerr << "Histogram sources." << endl;
	for (index_type i=0; i<N; ++i) {
	    if (trans->getChanReady(i)) continue;
	    chan_index ichn = trans->chan(i).index();
	    if (trans_online) {
		sl.bump(Channel_List[ichn], 0);
	    } else {
		sl.bump(Channel_List[ichn], trans->getChanFill(i));
	    }
	}

	//------------------------------  No more sources, mark the rest
	//                                of the channels as unavailable
	if (sl.empty()) {
	    if (debug(dlvl_meta)) cerr << "No more sources." << endl;
	    for (index_type i=0; i<N; ++i) {
		if (!trans->getChanReady(i)) {
		    trans->chan(i).set_stat(-1);
		}
	    }
	    break;
	}

	//------------------------------  save the best source, release the rest
	source_info si("");
	si.swap(sl.first());
	sl.clear();

	//------------------------------  Build channel list for best source
	driver_api::channel_vect chans;
	gps_type stop=trans->getStrideEnd();
	gps_type start=stop;
	gps_type tgps=trans->getStartGPS();
	string chStat;
	if (debug(dlvl_meta)) cerr << "Channels from source: " << si._name;
	for (index_type i=0; i<N; ++i) {
	    chan_index ichn = trans->chan(i).index();
	    if (trans->getChanReady(i)) {
		chStat = "ready";
	    } else if (Channel_List[ichn].is_source(si._name, tgps)) {
		chans.push_back(ichn);
		gps_type fill = trans->chan(i).fill();
		if (start > fill) start = fill;
		chStat = "fill";
	    } else {
		chStat = "not-here";
	    }
	    if (debug(dlvl_meta)) {
		cerr << ", " << Channel_List[ichn].name() << " " << chStat
		     << " " << start;
	    }
	}
	if (debug(dlvl_meta)) cerr << endl;

	//------------------------------  Get data from best source
	if (debug(dlvl_meta)) cerr << "Get data from: " << si._name << endl;
	const loc_list*  ll = si._loc;
	si.release();  // release source information
	const frame_loc* fl = ll->find_loc(start);
	frame_loc::driver_type drv = fl->driver();
	ll->unlock();
	int rc=TheDriverTable[drv].read_data(*ll, start, stop, chans);
	trans->reserve_segments();
	if (debug(dlvl_detail)) trans->dump(cerr);

	//------------------------------  Don't reuse the source.
	if (rc) sl.veto(si._name);
    }

    //---------------------------------- Log the number of iterations
    if (debug(dlvl_detail)) {
	ostringstream msg;
	msg << "number of iterations in fetch: " << iter_count << endl;
	debug_msg(dlvl_detail, trans->ident(), msg.str());
    }
}
